/*

  Simple C-like dynamic vectors/matrices

  use doble, no typedef no template

*/

#include <cstddef>
#include <iostream>

// if C++ < C++11 define nullptr
#if __cplusplus <= 199711L
  #include <cstdlib>
  #ifndef nullptr
    #include <cstddef>
    #define nullptr NULL
  #endif
#endif

/*
__     __        
\ \   / /__  ___ 
 \ \ / / _ \/ __|
  \ V /  __/ (__ 
   \_/ \___|\___|
*/
                 
using namespace std ;

// define a vector as a struct with the size of the vector
// and a pointer to its values

template <typename valueType>
class Vec {
  int         _size ;
  valueType * values ;

public:
  
  Vec() ; // default constructor
  Vec( int _size ) ; // another constructor

  ~Vec() ; // destructor

  Vec( Vec<valueType> const & ) ; // copy constructor
  Vec const & operator = ( Vec const & ) ; // copy operator

  template <typename T>
  Vec( Vec<T> const & a ) : values(nullptr), _size(0) {
    cout << "Copy constructor with conversion size = " << a.size() << "\n" ;
    allocate(a.size()) ;
    cout << "DONE\n" ;
    for ( int i = 0 ; i < _size ; ++i ) values[i] = a(i) ;  // inplicit conversion of type of a to the type of values
  }

  // copy operator
  template <typename T>
  Vec<valueType> const & 
  operator = ( Vec<T> const & a ) {
    cout << "Copy operator with conversion\n" ;
    allocate(a.size()) ;
    for ( int i = 0 ; i < _size ; ++i ) values[i] = a(i) ;
    return *this ;
  }

  int size() const { return _size ; }
  
  // methods
  void allocate( int size ) ;
  void fill( valueType val ) ;
  void fill_random() ;

  // magic methods, redefine operators!
  valueType
  operator () // name of the operator
  ( int i ) const // const is used to inform the compiler that operator do not modifty contents
  { return values[i] ; }   

  valueType & // return the reference (i.e. the address) of values[i]
  operator () ( int i )
  { return values[i] ; }   

} ;

/*
 __  __       _   
|  \/  | __ _| |_ 
| |\/| |/ _` | __|
| |  | | (_| | |_ 
|_|  |_|\__,_|\__|
*/

template <typename valueType>
class Mat {
  int          _nrow, _ncol ;
  valueType  * memory_block ; // allocated memory for the matrix
  valueType ** values ; // pointer to the rows of the matrix

public:

  Mat() ; // default constructor
  Mat( int _nr, int _nc ) ; // another constructor

  Mat( Mat const & ) ; // copy constructor
  Mat const & operator = ( Mat const & ) ; // copy operator

  ~Mat() ; // destructor

  int ncol() const { return _ncol ; }
  int nrow() const { return _nrow ; }
  
  void allocate( int nrow, int ncol ) ;
  void fill_random( ) ;
  void fill( valueType val ) ;

  // magic methods, redefine operators!
  valueType
  operator () // name of the operator
  ( int i, int j ) const // const is used to inform the compiler that operator do not modifty contents
  { return values[i][j] ; }

  valueType & // return the reference (i.e. the address) of values[i]
  operator () ( int i, int j )
  { return values[i][j] ; }

} ;

/*
 ___    _____  
|_ _|  / / _ \ 
 | |  / / | | |
 | | / /| |_| |
|___/_/  \___/ 

*/

template <typename valueType>
ostream &
operator << ( ostream & stream, Vec<valueType> const & vec ) ;

template <typename valueType>
ostream &
operator << ( ostream & stream, Mat<valueType> const & mat ) ;

/*
    _    _            _               
   / \  | | __ _  ___| |__  _ __ __ _ 
  / _ \ | |/ _` |/ _ \ '_ \| '__/ _` |
 / ___ \| | (_| |  __/ |_) | | | (_| |
/_/   \_\_|\__, |\___|_.__/|_|  \__,_|
           |___/                      
*/

template <typename valueType> Vec<valueType> operator + ( Vec<valueType> const & a, Vec<valueType> const & b ) ;
template <typename valueType> Vec<valueType> operator - ( Vec<valueType> const & a, Vec<valueType> const & b ) ;
template <typename valueType> Vec<valueType> operator * ( Vec<valueType> const & a, Vec<valueType> const & b ) ;
template <typename valueType> Vec<valueType> operator / ( Vec<valueType> const & a, Vec<valueType> const & b ) ;

template <typename valueType> Vec<valueType> operator * ( valueType a, Vec<valueType> const & b ) ;
template <typename valueType> Vec<valueType> operator * ( Vec<valueType> const & a, valueType b ) ;

template <typename valueType> Vec<valueType> operator / ( Vec<valueType> const & a, valueType b ) ;
template <typename valueType> Vec<valueType> operator ^ ( Vec<valueType> const & a, valueType b ) ;

template <typename valueType> valueType dot( Vec<valueType> const & a, Vec<valueType> const & b ) ;
template <typename valueType> valueType norm( Vec<valueType> const & a ) ;

// ------------

template <typename valueType> Mat<valueType> operator + ( Mat<valueType> const & A, Mat<valueType> const & B ) ;
template <typename valueType> Mat<valueType> operator - ( Mat<valueType> const & A, Mat<valueType> const & B ) ;
template <typename valueType> Mat<valueType> operator * ( Mat<valueType> const & A, Mat<valueType> const & B ) ;

template <typename valueType> Mat<valueType> operator * ( valueType a, Mat<valueType> const & B ) ;
template <typename valueType> Mat<valueType> operator * ( Mat<valueType> const & A, valueType b ) ;

template <typename valueType> Mat<valueType> operator / ( Mat<valueType> const & A, valueType b ) ;

// ------------
template <typename valueType> Vec<valueType> operator * ( Mat<valueType> const & A, Vec<valueType> const & b ) ;
