#include "mv.hh"
#include <assert.h> // for assert
#include <cmath>
#include <stdlib.h> // for rand() function

// initialize structure to a vector of size 0
template <typename valueType>
Vec<valueType>::Vec() : values(nullptr), _size(0) {
  cout << "Vec: Call constructor for Vec\n" ;
}

template <typename valueType>
Vec<valueType>::Vec( int _sz ) : _size(_sz) {
  cout << "Vec: Call constructor for Vec with size = " << _size << "\n" ;
  values = new valueType[_size] ; // allocate memory for the vector
}

// copy constructor
template <typename valueType>
Vec<valueType>::Vec( Vec const & a ) {
  cout << "Vec: Call copy constructor\n" ;
  allocate(a.size()) ;
  for ( int i = 0 ; i < _size ; ++i ) values[i] = a(i) ;
}

// copy operator
template <typename valueType>
Vec<valueType> const & 
Vec<valueType>::operator = ( Vec const & a ) {
  cout << "Vec: Call copy operator\n" ;
  allocate(a.size()) ;
  for ( int i = 0 ; i < _size ; ++i ) values[i] = a(i) ;
  return *this ;
}

template <typename valueType>
Vec<valueType>::~Vec( ) {
  cout << "Vec: Call detructor for Vec\n" ;
  // check if alreay de-allocated
  if ( values != nullptr ) delete [] values ;
  _size = 0 ;
}

// alocate a vector to teh size: size
template <typename valueType>
void
Vec<valueType>::allocate( int size ) {
  _size = size ;
  // check if alreay allocated
  if ( values != nullptr ) delete [] values ;
  values = new valueType[_size] ;
}

template <typename valueType>
void
Vec<valueType>::fill( valueType val ) {
  for ( int i = 0 ; i < _size ; ++i )
    values[i] = val ;
}

template <typename valueType>
void
Vec<valueType>::fill_random() {
  for ( int i = 0 ; i < _size ; ++i )
    values[i] = std::floor(100.0*((rand() - rand())/double(RAND_MAX))) ;
}

/*
 ___    _____  
|_ _|  / / _ \ 
 | |  / / | | |
 | | / /| |_| |
|___/_/  \___/ 

*/

template <typename valueType>
ostream &
operator << ( ostream & stream, Vec<valueType> const & vec ) {
  stream << vec(0) ;
  for ( int i = 1 ; i < vec.size() ; ++i )
    stream << " " << vec(i) ;
  stream << '\n' ;
  return stream ;
}

/*
    _    _            _               
   / \  | | __ _  ___| |__  _ __ __ _ 
  / _ \ | |/ _` |/ _ \ '_ \| '__/ _` |
 / ___ \| | (_| |  __/ |_) | | | (_| |
/_/   \_\_|\__, |\___|_.__/|_|  \__,_|
           |___/                      
*/

template <typename valueType>
Vec<valueType>
operator + ( Vec<valueType> const & a, Vec<valueType> const & b ) {
  assert( a.size() == b.size() ) ;
  Vec<valueType> res(a.size()) ;
  for ( int i = 0 ; i < a.size() ; ++i ) res(i) = a(i) + b(i) ;
  return res ;
}

template <typename valueType>
Vec<valueType>
operator - ( Vec<valueType> const & a, Vec<valueType> const & b ) {
  assert( a.size() == b.size() ) ;
  Vec<valueType> res(a.size()) ;
  for ( int i = 0 ; i < a.size() ; ++i ) res(i) = a(i) - b(i) ;
  return res ;
}

template <typename valueType>
Vec<valueType>
operator * ( Vec<valueType> const & a, Vec<valueType> const & b ) {
  assert( a.size() == b.size() ) ;
  Vec<valueType> res(a.size()) ;
  for ( int i = 0 ; i < a.size() ; ++i ) res(i) = a(i) * b(i) ;
  return res ;
}

template <typename valueType>
Vec<valueType>
operator / ( Vec<valueType> const & a, Vec<valueType> const & b ) {
  assert( a.size() == b.size() ) ;
  Vec<valueType> res(a.size()) ;
  for ( int i = 0 ; i < a.size() ; ++i ) res(i) = a(i) / b(i) ;
  return res ;
}

template <typename valueType>
Vec<valueType>
operator * ( valueType a, Vec<valueType> const & b ) {
  Vec<valueType> res(b.size()) ;
  for ( int i = 0 ; i < b.size() ; ++i ) res(i) = a*b(i) ;
  return res ;
}

template <typename valueType>
Vec<valueType>
operator * ( Vec<valueType> const & a, valueType b ) {
  Vec<valueType> res(a.size()) ;
  for ( int i = 0 ; i < a.size() ; ++i ) res(i) = a(i)*b ;
  return res ;
}

template <typename valueType>
Vec<valueType>
operator / ( Vec<valueType> const & a, valueType b ) {
  Vec<valueType> res(a.size()) ;
  for ( int i = 0 ; i < a.size() ; ++i ) res(i) = a(i)/b ;
  return res ;
}

template <typename valueType>
Vec<valueType>
operator ^ ( Vec<valueType> const & a, valueType b ) {
  Vec<valueType> res(a.size()) ;
  for ( int i = 0 ; i < a.size() ; ++i ) res(i) = std::pow(a(i),b) ;
  return res ;
}

template <typename valueType>
valueType
dot( Vec<valueType> const & a, Vec<valueType> const & b ) {
  assert( a.size() == b.size() ) ;
  double res(0) ;
  for ( int i = 0 ; i < a.size() ; ++i ) res += a(i)*b(i) ;
  return res ;
}

template <typename valueType>
valueType norm( Vec<valueType> const & a ) { return sqrt(dot(a,a)) ; }

template class Vec<float> ; // force instantiation of all the methods of class Mat<double>
template class Vec<double> ; // force instantiation of all the methods of class Mat<double>
template class Vec<int> ; // force instantiation of all the methods of class Mat<double>

template ostream & operator << ( ostream & stream, Vec<float> const & mat ) ;
template ostream & operator << ( ostream & stream, Vec<double> const & mat ) ;
template ostream & operator << ( ostream & stream, Vec<int> const & mat ) ;

template Vec<float> operator + ( Vec<float> const & a, Vec<float> const & b ) ;
template Vec<float> operator - ( Vec<float> const & a, Vec<float> const & b ) ;
template Vec<float> operator * ( Vec<float> const & a, Vec<float> const & b ) ;
template Vec<float> operator / ( Vec<float> const & a, Vec<float> const & b ) ;
template Vec<float> operator * ( float a, Vec<float> const & b ) ;
template Vec<float> operator * ( Vec<float> const & a, float b ) ;
template Vec<float> operator / ( Vec<float> const & a, float b ) ;
template Vec<float> operator ^ ( Vec<float> const & a, float b ) ;
template float      dot( Vec<float> const & a, Vec<float> const & b ) ;

template Vec<double> operator + ( Vec<double> const & a, Vec<double> const & b ) ;
template Vec<double> operator - ( Vec<double> const & a, Vec<double> const & b ) ;
template Vec<double> operator * ( Vec<double> const & a, Vec<double> const & b ) ;
template Vec<double> operator / ( Vec<double> const & a, Vec<double> const & b ) ;
template Vec<double> operator * ( double a, Vec<double> const & b ) ;
template Vec<double> operator * ( Vec<double> const & a, double b ) ;
template Vec<double> operator / ( Vec<double> const & a, double b ) ;
template Vec<double> operator ^ ( Vec<double> const & a, double b ) ;
template double      dot( Vec<double> const & a, Vec<double> const & b ) ;

template Vec<int> operator + ( Vec<int> const & a, Vec<int> const & b ) ;
template Vec<int> operator - ( Vec<int> const & a, Vec<int> const & b ) ;
template Vec<int> operator * ( Vec<int> const & a, Vec<int> const & b ) ;
template Vec<int> operator / ( Vec<int> const & a, Vec<int> const & b ) ;
template Vec<int> operator * ( int a, Vec<int> const & b ) ;
template Vec<int> operator * ( Vec<int> const & a, int b ) ;
template Vec<int> operator / ( Vec<int> const & a, int b ) ;
template Vec<int> operator ^ ( Vec<int> const & a, int b ) ;
template int dot( Vec<int> const & a, Vec<int> const & b ) ;

